import requests
import json
import random
import os
import datetime


class User:
    """
    This is the main User class.

    Parameters
    ----------
    id_hash: int
        An ID hash of the dataset in the API.
    attributes: dic
        A dictionary holding the attributes of a dataset.
    fname: str
        A path/file string pointing to the location of a file for upload.
    sever: str
        A URL string of the vizzuality server.
    """

    def __init__(self):
        self.home = os.path.expanduser('~')
        self.hidden_folder_path = f"{self.home}/.Skydipper"
        self.hidden_creds_file_path = f"{self.hidden_folder_path}/creds"
        try:
            os.stat(self.hidden_folder_path) #<-- Does the .Skydipper credential folder exist?
            #print(f"Hidden credential folder found at {self.hidden_folder_path}")
        except:
            os.mkdir(self.hidden_folder_path) # If not, create the folder in a users home dir
            #print(f"No hidden folder - generating one at {self.hidden_folder_path}")
        # At this point, a hidden folder definitley exists
        # Now it's time to get some authorization data...
        # First see if there is a credential file
        if not os.path.exists(self.hidden_creds_file_path):
            #print("No token file - Generating one now")
            self.gen_token()
        # At this point, there should be a cred file. Read the credential from the file and Test if it is valid
        self.token = self.read_token()
        if not self.token_valid():
            self.gen_token()
        #print(f"Authentication success: {self.token_valid()}")
        self.headers =  {
                        'Authorization': f'Bearer {self.token}',
                        'Content-Type': 'application/json',
                        'Cache-Control': 'no-cache'
                        }

    def gen_token(self):
        """
            Use the authorization endpoint to go to https://api.skydipper.com/auth
        """
        email = input("Please enter your the email address associated with your Skydipper account")
        password = input("Please enter your Skydipper password")
        payload = {
            "email": email,
            "password": password
        }
        url = 'https://api.skydipper.com/auth/login'
        headers = {'Content-Type': 'application/json'}
        r = requests.post(url, json=payload, headers=headers)
        if r.status_code == 200:
            self.user_id = r.json().get('data').get('id')
            self.createdAt = r.json().get('data').get('createdAt')
            self.role = r.json().get('data').get('role')
            self.extraUserData = r.json().get('data').get('extraUserData')
            self.token = r.json().get('data').get('token')
            self.save_creds()
        return

    def save_creds(self):
        with open(self.hidden_creds_file_path, 'w') as opened_file:
            opened_file.write(self.token)

    def read_token(self):
        """Read the token from the creds file"""
        with open(self.hidden_creds_file_path, 'r') as opened_file:
            tmp = opened_file.readlines()[0]
        return tmp

    def token_valid(self):
        url = "https://api.skydipper.com/api/v1/microservice"
        headers = {
            'Authorization': f'Bearer {self.token}',
            'Content-Type': 'application/json'
        }
        r = requests.get(url, headers=headers)
        if r.status_code == 200:
            return True
        else:
            return False



import requests
import folium
import urllib
import json
import random
import geopandas as gpd
from shapely.geometry import shape
import shapely.wkt
import geojson
from .utils import html_box, get_geojson_string
import json
from .user import User

class Geometry:
    """
    This is the main geometry class.

    Parameters
    ----------
    id_hash: int
        An ID hash of a Geostore from a Vizzuality endpoint.
    attributes: dic
        A valid geojson representation.
    server: str
        The string of the server URL.
    s: obj
        A shapely object.
    """
    def __init__(self, id_hash=None, attributes=None, s=None, parameters=None, server='https://api.skydipper.com'):
        self.server = server
        self.User = User()
        if s:
            attributes = self.create_attributes_from_shapely(s)
        if parameters:
            id_hash = self.get_id_from_params(parameters)
        if attributes:
            self.attributes = self.create_geostore_from_geojson(attributes)
        else:
            self.id = id_hash
            self.attributes = self.get_geometry()

    def __repr__(self):
        return self.__str__()

    def __str__(self):
        return f"Geometry {self.id}"

    def _repr_html_(self):
        return html_box(item=self)

    def create_attributes_from_shapely(self, s):
        """Using a Shapely object, we should build a Geometry object."""
        if s.geom_type in ['Polygon', 'Point', 'MultiPoint','MultiPolygon']:
            atts={'geojson': {'type': 'FeatureCollection',
                            'features': [{'type': 'Feature',
                                'properties': {},
                                'geometry': geojson.Feature(geometry=s, properties={}).get('geometry')
                                        }]}}
            return atts
        else:
            raise ValueError('shape object was not of suitable geometry type')

    def get_id_from_params(self, parameters):
        """
        """
        raise ValueError(f'Params functionality not supported in Skydipper.')

    def create_geostore_from_geojson(self, attributes):
        """Parse valid geojson into a geostore object and register it to a
        Gestore object on a server. Return the object, and instantiate a
        Geometry.
        """
        try:
            body = json.loads(json.dumps(attributes))
        except:
            raise ValueError(f"Unable to pass attributes. Expected valid geojson, recieved: {attributes}")
        url = self.server + '/v1/geostore'
        r = requests.post(url, headers=self.User.headers, json=body)
        if r.status_code == 200:
            self.id = r.json().get('data').get('id')
            return r.json().get('data').get('attributes')
        else:
            raise ValueError(f'Recieved response of {r.status_code} from {r.url} when posting to geostore.')

    def get_geometry(self, simplify=False, version='v2'):
        """
        Returns a geostore object by ID from a Vizzuality endpoint.
        """
        hash = random.getrandbits(16)
        url = (f'{self.server}/{version}/geostore/{self.id}?simplify={simplify}&hash={hash}')
        r = requests.get(url, headers=self.User.headers)
        if r.status_code == 200:
            return r.json().get('data').get('attributes')
        else:
            raise ValueError(f'Unable to get dataset {self.id} from {r.url}')

    def table(self):
        """
        Returns features as GeoDataFrame
        """
        attributes = self.attributes
        props = {
            **attributes['info'],
            'areaHa': attributes['areaHa'],
            'bbox': attributes['bbox'],
            'id': attributes['hash']
         }
        features = attributes['geojson']['features']
        if len(features) > 0:
            gdf = gpd.GeoDataFrame([{**props, 'geometry': shape(feature['geometry'])} for feature in features]).set_geometry('geometry')
            gdf.crs = {'init' :'epsg:4326'}
            return gdf
        return []

    def shape(self):
        """
        Returns features as a list of Shapely geometries
        """
        features = self.attributes['geojson']['features']
        if len(features) > 0:
            return [shape(feature['geometry']) for feature in features]
        else:
            return None

    @staticmethod
    def get_image_url(centroid, band_viz, start, end):
        params = {"lon": centroid[1],
                  "lat": centroid[0],
                  "start": start,
                  "end": end,
                  "band_viz": json.dumps(band_viz)
                  }
        url = "https://api.skydipper.com/v1/recent-imagery"
        r = requests.get(url, params=params, headers=self.User.headers)
        if r.status_code == 200:
            tile_url = r.json().get('data').get('tiles')[0].get('attributes').get('tile_url')
            return tile_url
        else:
            return None

    def get_composite_url(self, centroid, band_viz, instrument, date_range):
        valid_servers = ["https://api.skydipper.com"]
        if self.server in valid_servers:
            params = {"geostore": self.id,
                      "instrument": instrument,
                      "date_range": date_range,
                      "band_viz": json.dumps(band_viz)
                     }
            url = "/v1/composite-service"
            url = self.server + url
            r = requests.get(url, params=params, headers=self.User.headers)
            if r.status_code == 200:
                tile_url = r.json().get('attributes').get('tile_url')
                return tile_url
        else:
            url = "https://api.skydipper.com/v1/composite-service/geom"
            payload = json.dumps(self.attributes)
            params = {"instrument": instrument,
                      "date_range": date_range,
                      "band_viz": json.dumps(band_viz)
                     }
            r = requests.request("POST", url, data=payload, headers=self.User.headers, params=params)
            if r.status_code == 200:
                tile_url = r.json().get('attributes').get('tile_url')
                return tile_url

    def map(self, image=False, instrument='sentinel', start='2017-01-01', end='2018-01-01', color='#64D1B8', weight=4):
        """
        Returns a folium map with styles applied via attributes.

        Parameters
        ----------
        image: bool
            If true, a satellite image will be returned related to your geometry.
        instrument: str
            Either landsat or sentinel. Depending on the string a different satellite
            instrument will be used.
        start: str
            Start date for the composite time-range of the satellite image. If point
            data the best intersecting image will be returned within the specified
            time-range. If polygon-type data a cloud-free composite within the
            time-range will be returned.
        end: str
            End date for the composite time-range. If point
            data the best intersecting image will be returned within the specified
            time-range. If polygon-type data a cloud-free composite within the
            time-range will be returned.
        weight: int
            Weight of geom outline. Default = 4.
        color: str
            Hex code for geom outline. Default = #64D1B8.
        """
        if instrument == 'sentinel':
            band_viz = {'bands': ['B4', 'B3', 'B2'], 'min': 0, 'max': 0.4}
        else:
            band_viz = {'bands': ['B4', 'B3', 'B2'], 'min':0, 'max':0.2, 'gamma':[1.3, 1.3, 1.3]}
        geojson = self.attributes['geojson']
        geometry = geojson['features'][0]['geometry']
        bbox = self.attributes['bbox']
        centroid = list(self.shape()[0].centroid.coords)[0][::-1]
        bounds = [bbox[2:][::-1], bbox[:2][::-1]]
        result_map = folium.Map(location=centroid, tiles='OpenStreetMap', zoom_start=6)
        result_map.fit_bounds(bounds)
        if geometry['type'] == 'Point' or geometry['type'] == 'MultiPoint':
            folium.Marker(centroid).add_to(result_map)
            if image:
                tile_url = self.get_image_url(centroid=centroid, band_viz=band_viz,
                                              start=start, end=end)
                result_map.add_tile_layer(tiles=tile_url, attr=f"{instrument} image")
        else:
            if image:
                date_range = f'{start},{end}'
                tile_url = self.get_composite_url(centroid=centroid, band_viz=band_viz,
                                    instrument=instrument, date_range=date_range)
                result_map.add_tile_layer(tiles=tile_url, attr=f"{instrument} image")
            style_function = lambda x: {
                'fillOpacity': 0.0,
                    'weight': weight,
                    'color': color
                    }
            folium.GeoJson(data=get_geojson_string(geometry), style_function=style_function).add_to(result_map)
        return result_map

    def describe(self, lang='en', app='gfw'):
        """Returns an object with a title and description of a region. Running
        this function adds a property to the object of self.description.

        Parameters
        ----------
        lang: str
            A 2-character language string of the description output
        app: str
            An optional application id string (such as 'gfw') to tailor the description to.
        """
        # If the geostore exists on the right server, send the id, else send the valid geojson attributes
        valid_servers = ["https://api.skydipper.com"]
        if self.server in valid_servers:
            params = {"geostore": self.id,
                      "lang": lang,
                      "app": app}
            url = f"{self.server}/v1/geodescriber"
            r = requests.get(url, params=params, headers=self.User.headers)
            if r.status_code == 200:
                if self.server == 'https://api.skydipper.com':
                    tmp = r.json().get('attributes')
                else:
                    tmp = r.json().get('data')
                d = {'title': tmp.get('title'),
                    'description': tmp.get('description'),
                    'lang': tmp.get('lang')}
                self.description = d
                print(f"Title: {d.get('title')}")
                return
            else:
                print(f"Description attempt failed: response: {r.status_code}, \n {r.json()}")
                return None
        else:
            url = "https://api.skydipper.com/v1/geodescriber/geom"
            payload = json.dumps(self.attributes)
            querystring = {"lang": lang, "app": app}
            r = requests.request("POST", url, data=payload, headers=self.User.headers, params=querystring)
            if r.status_code == 200:
                d = {'title': r.json().get('data').get('title'),
                     'description': r.json().get('data').get('description'),
                     'lang': r.json().get('data').get('lang')}
                self.description = d
                print(f"Title: {d.get('title')}")
                print(f"Description: {d.get('description')}")
                return
            else:
                print(f"Description attempt failed: response: {r.status_code}, \n {r.json()}")
                return None
        return None

    def simplify(self, tolerance, preserve_topology=True):
        """
        Returns a simplified geometry produced by the Douglas-Peucker
        algorithm. Coordinates of the simplified geometry will be no more than the
        tolerance distance from the original. Unless the topology preserving
        option is used, the algorithm may produce self-intersecting or
        otherwise invalid geometries. Returns a Geostore object.

        Parameters
        ----------
        tolerance: int
            Tolerance value
        preserve_topology: bool
            Should the topology be presevered, True or False.
        """
        s = self.shape()[0]
        tmp_s = s.simplify(tolerance=tolerance, preserve_topology=True)
        return Geometry(s=tmp_s)


import requests
import json
from .image import Image
from .utils import create_class, show_image_collection, flatten_list

class ImageCollection:
    """
    Returns a list of Image objects based on search criteria.

    This function searches all available images in sentinel and landsat
    and returns an image.

    Parameters
    ----------
    lat: float
        A decimal latitude.

    lon: float
        A decimal longitude.

    band_viz: dic
        A dictionary of bands and visulisation (e.g. {'bands': ['B4', 'B3', 'B2'], 'min': 0, 'max': 0.4}).

    instrument: str
        A string indicating the satellite platform ('sentinel', 'landsat', 'all').
        Note - instrument filtering is not yet implemented.

    start: str
        Start date ('YYYY-MM-DD') to bound the search for the satellite images.

    end: str
        End date ('YYYY-MM-DD') to bound the search for the satellite images.
    """

    def __init__(self, lat, lon, start, end, instrument=None,
                    band_viz={'bands': ['B4', 'B3', 'B2'], 'min': 0, 'max': 0.4},
                    server='https://production-api.globalforestwatch.org'):
        self.lat = lat
        self.lon = lon
        self.start = start
        self.end = end
        self.instrument = instrument
        self.band_viz = band_viz
        self.server = server
        self.collection = self.get_collection()
        self.iter_position = 0

    def _repr_html_(self):
        str_html = ""
        for n, c in enumerate(self.collection):
            str_html += show_image_collection(c, n)
            if n < len(self.collection)-1:
                str_html += '<p></p>'
        return str_html

    def __repr__(self):
        rep_string = "["
        for n, c in enumerate(self.collection):
            rep_string += str(f"{n}. {c.get('type')} {c.get('instrument')} {c.get('date_time')}")
            if n < len(self.collection)-1:
                rep_string += ',\n '
        rep_string += ']'
        return rep_string

    def __iter__(self):
        return self

    def __next__(self):
        if self.iter_position >= len(self.collection):
            self.iter_position = 0
            raise StopIteration
        else:
            self.iter_position += 1
            return self.collection[self.iter_position - 1]

    def __getitem__(self, key):
        items = self.collection[key]
        if type(items) == list:
            return [create_class(item) for item in items]
        else:
            return create_class(items)

    def __len__(self):
        return len(self.collection)

    def get_collection(self):
        """Returns a list of Image objects."""
        url = 'https://production-api.globalforestwatch.org/recent-tiles'
        params = {'lon':self.lon,
                  'lat':self.lat,
                  'start':self.start,
                  'end':self.end}
        r = requests.get(url=url, params=params)
        if(r.status_code != 200):
            raise ValueError(f'Bad response from recent-tiles service: {r.status_code}, {r.json()}')
        try:
            image_list = []
            for item in r.json().get('data').get('tiles'):
                tmp = {'instrument': item.get('attributes').get('instrument'),
                        'date_time': item.get('attributes').get('date_time'),
                        'cloud_score': item.get('attributes').get('cloud_score'),
                        'source': item.get('attributes').get('source'),
                        'band_viz': self.band_viz,
                        'server': self.server,
                        'thumb_url': None,
                        'type': 'Image',
                        'bbox': item.get('attributes').get('bbox')
                    }
                image_list.append(tmp)
            # At this point we need to do a batch request for the image_thumbnails and add that key
            source_list = [{'source': item.get('source')} for item in image_list]
            payload = {'source_data': source_list, 'bands': self.band_viz.get('bands')}
            url = self.server + '/recent-tiles/thumbs'
            r2 = requests.post(url, data=json.dumps(payload), headers={'Content-Type': 'application/json'})
            if r2.status_code == 200:
                for n, item in enumerate(r2.json().get('data').get('attributes')):
                    image_list[n]['thumb_url'] = item.get('thumbnail_url')
            else:
                raise ValueError(f'Response from image thumbnail service was {r2.status_code}: {r2.json()}')
            return image_list
        except:
            raise ValueError(f'Failed attempting to work with {item}')

    @staticmethod
    def ring_list_to_json_bbox(bbox):
        """If a bbox has been returned without all the trimmings, as in the case of
        the cloud func microservices we add some extra structure and return it.
        """
        tmp = [[bbox[0],bbox[1]],
               [bbox[0], bbox[3]],
               [bbox[2], bbox[3]],
               [bbox[2], bbox[1]],
               [bbox[0], bbox[1]]
              ]
        return {'geometry': {'coordinates': tmp, 'type': 'Polygon'}}

    def composite(self, instrument='Sentinel2'):
        """
        Make a composite focused on a region for a specific Instrument. Returns an Image object.

        Parameters
        ----------
        instrument: str
            A string describing the Instrument type to composite ('Sentinel2', 'Landsat')
        """
        payload = {'instrument': instrument,
                  'lon': self.lat,                 # tep fix for a bug in MS
                  'lat': self.lon,
                  'start': self.start,
                  'end': self.end}
        url = f'https://us-central1-skydipper-196010.cloudfunctions.net/composite'
        headers = {'Content-Type': 'application/json'}
        r = requests.post(url, data=json.dumps(payload), headers=headers)
        if r.status_code == 200:
            tmp = {'instrument': instrument,
                    'date_time': f'{self.start}–{self.end}',
                    'cloud_score': '-',
                    'source': instrument,
                    'band_viz': self.band_viz,
                    'server': self.server,
                    'thumb_url': r.json().get('thumb_url'),
                    'tile_url': r.json().get('tile_url'),
                    'type': 'Composite Image',
                    'bbox': self.ring_list_to_json_bbox(r.json().get('bbox')),
                    'np_array_bounds':r.json().get('bbox')
                        }
            return Image(**tmp)
        else:
            raise ValueError(f"Failure in composite service: {r.status_code} {r.url}")



import requests
import random
import json
from .utils import html_box, nested_set
from .user import User

class Metadata:
    """
    This is the main Metadata class.

    Parameters
    ----------
    attributes: dic
        A dictionary holding the attributes of a metadata (which are attached to a Dataset).
    """
    def __init__(self, attributes=None, server="https://api.skydipper.com"):
        if attributes.get('type') != 'metadata':
            raise ValueError(f"Non metadata attributes passed to Metadata class ({attributes.get('type')})")
        self.id = attributes.get('id')
        self.server = server
        self.type = "Metadata"
        self.User = User()
        self.attributes = attributes.get('attributes')

    def __repr__(self):
        return self.__str__()

    def __str__(self):
        return f"Metadata {self.id}"

    def update(self, update_params=None, token=None):
        """
        Update the attributes of a Metadata object providing a RW-API token is supplied.

        A single application string and language string ('en' by default) must be specified within the
        `update_params` dictionary, as well as an (optional) info dictionary.
        Info has a free schema.
        """
        from .dataset import Dataset
        if not token:
            raise ValueError(f'[token] API token required to update metadata.')
        app = self.attributes.get('application', None)
        lang = update_params.get('language', 'en')
        info = update_params.get('info', None)
        ds_id = self.attributes.get('dataset', None)
        if info and app:
            payload = {
                "application": app,
                "language": lang,
                "info": info,
            }
            print('payload', payload)
            try:
                url = f'{self.server}/v1/dataset/{ds_id}/metadata'
                print('url',url)
                r = requests.patch(url, data=json.dumps(payload), headers=self.User.headers)
            except:
                raise ValueError(f'Metadata update failed.')
            if r.status_code == 200:
                print(f'Metadata updated.')
                return Dataset(id_hash=ds_id, server=self.server).metadata
            else:
                print(f'Failed with error code {r.status_code}')
                return None
        else:
            raise ValueError(f'Metadata update requires info object and application string.')

    def delete(self, token=None):
        """
        Delete the current metadata, removing it's association to the parent dataset.
        A RW-API token is required.
        """
        if not token:
            raise ValueError(f'[token] API token required to delete vocabulary.')
        lang = self.attributes.get('language', None)
        app = self.attributes.get('application', None)
        ds_id = self.attributes.get('dataset', None)
        if lang and app:
            try:
                url = f'{self.server}/dataset/{ds_id}/metadata?application={app}&language={lang}'
                r = requests.delete(url, headers=self.User.headers)
            except:
                raise ValueError(f'Metdata deletion failed.')
            if r.status_code == 200:
                print(f'Metdata deleted.')
        return None

class Vocabulary:
    """
    This is the main Vocabulary class.

    Parameters
    ----------
    attributes: dic
        A dictionary holding the attributes of a vocabulary (which are attached to a Dataset).
    """
    def __init__(self, attributes=None, server="https://api.skydipper.com"):
        if attributes.get('type') != 'vocabulary':
            raise ValueError(f"Non vocabulary attributes passed to Vocabulary class ({attributes.get('type')})")
        self.server = server
        self.attributes = attributes.get('attributes')
        self.id = self.attributes.get('resource').get('id')

    def __repr__(self):
        return self.__str__()

    def __str__(self):
        return f"Vocabulary {self.id}"

    def update(self, update_params=None, token=None):
        """
        Update the attributes of a Vocabulary object providing a RW-API token is supplied.

        A single application string, name string and tags list must be specified within the `update_params` dictionary.
        """
        from .dataset import Dataset
        if not token:
            raise ValueError(f'[token] API token required to update vocabulary.')
        update_params['application'] = self.attributes.get('application', None)
        ds_id = self.id
        self.delete(token=token)
        Dataset(id_hash=ds_id, server=self.server).add_vocabulary(vocab_params=update_params, token=token)
        return Dataset(id_hash=ds_id, server=self.server).vocabulary

    def delete(self, token=None):
        """
        Delete the current vocabulary, removing it's association to the parent dataset.
        A RW-API token is required.
        """
        if not token:
            raise ValueError(f'[token] API token required to delete vocabulary.')
        vocab_type = self.attributes.get('name', None)
        app = self.attributes.get('application', None)
        ds_id = self.id
        if vocab_type and app:
            try:
                url = f'{self.server}/dataset/{ds_id}/vocabulary/{vocab_type}?app={app}'
                headers = {'Authorization': 'Bearer ' + token, 'Content-Type': 'application/json', 'Cache-Control': 'no-cache'}
                r = requests.delete(url, headers=headers)
            except:
                raise ValueError(f'Vocabulary deletion failed.')
            if r.status_code == 200:
                print(f'Vocabulary {vocab_type} deleted.')
        return None

class Widget:
    """
    This is the main Widget class.

    Parameters
    ----------
    attributes: dic
        A dictionary holding the attributes of a widget (which are attached to a Dataset).
    """
    def __init__(self, id_hash=None, attributes=None, server="https://api.skydipper.com"):
        self.id = id_hash
        self.server = server
        if id_hash:
            self.attributes = self.get_widget()
        elif attributes:
            self.id = attributes.get('id')
            self.attributes = self.get_widget()

        else:
            raise ValueError(f'Unable to initialise Widget without id_hash.')

    def __repr__(self):
        return self.__str__()

    def _repr_html_(self):
        return html_box(item=self)

    def __str__(self):
        return f"Widget {self.id} {self.attributes.get('name','')}"

    def get_widget(self):
        """
        Returns a widget from a Vizzuality API.
        """
        try:
            hash = random.getrandbits(16)
            url = (f'{self.server}/v1/widget/{self.id}?hash={hash}')
            r = requests.get(url)
        except:
            raise ValueError(f'Unable to get Widget {self.id} from {r.url}')

        if r.status_code == 200:
            return r.json().get('data').get('attributes')
        else:
            raise ValueError(f'Widget with id={self.id} does not exist.')

    def update(self, update_params=None, token=None):
        """
        Update the attributes of a Widget object providing a RW-API token is supplied.

        A single application string must be specified within the
        `update_params` dictionary, as well as an (optional) widgetCOnfig dictionary.

        Note, widgetConfig has a free schema.
        """
        from .dataset import Dataset
        if not token:
            raise ValueError(f'[token] API token required to update widget.')
        ds_id = self.attributes.get('dataset', None)
        w_id = self.id
        update_keys = ["widgetConfig", "name", "description", "application", "default", "protected", "defaultEditableWidget", "published", "freeze"]
        attributes = {f'{k}':v for k,v in self.attributes.items() if k in update_keys}
        if update_params and any([x in update_keys for x in list(update_params.keys())]):
            payload = {}
            for k, v in update_params.items():
                if '.' in k:
                    nested_keys = k.split('.')
                    if len(nested_keys) > 1 and nested_keys[0] in list(attributes.keys()):
                        payload[nested_keys[0]] = attributes.get(nested_keys[0])
                        nested_set(payload, nested_keys, v)
                elif k in list(attributes.keys()):
                    payload[k] = v
            try:
                url = f'{self.server}/v1/dataset/{ds_id}/widget/{w_id}'
                print('url',url)
                headers = {'Authorization': 'Bearer ' + token, 'Content-Type': 'application/json'}
                r = requests.patch(url, data=json.dumps(payload), headers=headers)
            except:
                raise ValueError(f'Widget update failed.')
            if r.status_code == 200:
                print(f'Widget updated.')
                self.attributes = self.get_widget()
                return self
            else:
                print(f'Failed with error code {r.status_code}')
                return None
        else:
            raise ValueError(f'Widget update requires update_params object.')

    def delete(self, token=None):
        """
        Delete the current widget.
        A RW-API token is required.
        """
        if not token:
            raise ValueError(f'[token] API token required to delete vocabulary.')
        w_id = self.id
        ds_id = self.attributes.get('dataset', None)
        try:
            url = f'{self.server}/dataset/{ds_id}/widget/{w_id}'
            headers = {'Authorization': 'Bearer ' + token, 'Content-Type': 'application/json', 'Cache-Control': 'no-cache'}
            r = requests.delete(url, headers=headers)
        except:
            raise ValueError(f'Widget deletion failed.')
        if r.status_code == 200:
            print(f'Widget deleted.')
        return None

    def save(self, path=None):
        """
        Construct dataset json and save to local path in a date-referenced folder
        """
        from .dataset import Dataset
        ds_id = self.attributes['dataset']
        ds = Dataset(id_hash=ds_id, server=self.server)
        ds.save(path=path)

    def merge(self, token=None, target_widget=None, target_widget_id=None, target_server="https://api.skydipper.com", key_whitelist=[], force=False):
        """
        'Merge' one Widget entity into another target Widget.
        The argument `key_whitelist` can be used to specify which properties you wish to merge (if not all)
        Note: requires API token.
        """
        if not token:
            raise ValueError(f'[token] API token required to update Dataset.')
        if not target_widget and target_widget_id and target_server:
            target_widget = Widget(target_widget_id, server=target_server)
        else:
            raise ValueError(f'Requires either target Dataset or Dataset id plus server.')
        atts = self.attributes
        payload = {
            'widgetConfig': atts.get('widgetConfig', None),
            'connectorUrl': atts.get('connectorUrl', None),
            'name': atts.get('name', None),
            'description': atts.get('description', None),
            'application': atts.get('application', None),
        }
        if not key_whitelist: key_whitelist = [k for k in payload.keys()]
        filtered_payload = {k:v for k,v in payload.items() if v and k in key_whitelist}
        print(f'Merging {self.id} from {self.server} into {target_widget_id} on {target_server}.\nAre you sure you sure you want to continue?')
        if not force:
            conf = input()
        else:
            conf = 'y'
        if conf.lower() == 'y':
            try:
                merged_widget = target_widget.update(update_params=filtered_payload, token=token)
            except:
                print('Aborting...')
            print('Completed!')
            return merged_widget

        elif conf.lower() == 'n':
            print('Aborting...')
            return False
        else:
            print('Requires y/n input!')
            return False


import json
import math
import ee
from time import sleep
from google.cloud import storage

def html_box(item):
    """Returns an HTML block with template strings filled-in based on item attributes."""
    is_layer = str(type(item)) == "<class 'Skydipper.layer.Layer'>"
    is_dataset = str(type(item)) == "<class 'Skydipper.dataset.Dataset'>"
    is_widget = str(type(item)) == "<class 'Skydipper.Skydipper.Widget'>"
    is_geometry = str(type(item)) == "<class 'Skydipper.geometry.Geometry'>"
    is_image = str(type(item)) == "<class 'Skydipper.image.Image'>"
    site_link = "<a href='https://skydipper.com/' target='_blank'>"
    site_logo = "<img class='itemThumbnail' src='https://skydipper.com/images/logo.png'>"
    if is_layer:
        kind_of_item = 'Layer'
        url_link = f'{item.server}/v1/layer/{item.id}?includes=metadata'
    elif is_dataset:
        kind_of_item = 'Dataset'
        url_link = f'{item.server}/v1/dataset/{item.id}?includes=metadata,layer'
    elif is_image:
        if item.type in ['Classified Image', 'Composite Image']:
            instrument = item.type
        else:
            instrument = item.instrument
        html_string = ("<div class='item_container' style='height: auto; overflow: hidden; border: 1px solid #2BA4A0;"
                    "border-radius: 5px; background: #2BA4A0; line-height: 1.21429em; padding: 10px;''>"
                    "<div class='item_left' style='width: 100px; height: 100px; float: left;''>"
                    f"<a href='{item.thumb_url}' target='_blank'>"
                    f"<img class='itemThumbnail' src='{item.thumb_url}'>"
                    "</a></div><div class='item_right' style='float: none; width: auto; hidden;padding-left: 10px; overflow: hidden;''>"
                    f"<b>Image Source</b>: {instrument} </br>"
                    f"<b>Datetime</b>: {item.date_time} </br>"
                    f"<b>Cloud score </b>: {item.cloud_score} </br>"
                    " </div> </div>")
        return html_string
    elif is_geometry:
        kind_of_item = 'Geometry'
        url_link = f'{item.server}/v1/geostore/{item.id}'
        html_string = ("<div class='item_container' style='height: auto; overflow: hidden; border: 1px solid #2bA4A0;"
            "border-radius: 5px; background: #2bA4A0; line-height: 1.21429em; padding: 10px;''>"
            "<div class='item_left' style='width: 210px; float: left;''>"
            f"{site_link}"
            f"{site_logo}"
            "</a></div><div class='item_right' style='float: none; width: auto; hidden;padding-left: 10px; overflow: hidden;''>"
            f"<b>Geometry id</b>: <a href={url_link} target='_blank'>{item.id}</a></br>")
        for k,v in item.attributes.get('info').items():
            if v and k != 'simplifyThresh':
                html_string += f"<br><i>{k}: {v}</i>"
        html_string += (""
            " </div> </div>")
        return html_string
    else:
        kind_of_item = 'Unknown'
        url_link = None
    table_statement = f"Data source {item.attributes.get('provider')}"
    if item.attributes.get('connectorUrl') and item.attributes.get('provider') == "cartodb":
        table_statement = (f"Carto table: <a href={item.attributes.get('connectorUrl')}"
                           " target='_blank'>"
                           f"{item.attributes.get('tableName')}"
                           "</a>"
                          )
    if item.attributes.get('provider') == 'gee':
        table_statement = (f"GEE asset: <a href='https://code.earthengine.google.com/asset='"
                           f"{item.attributes.get('tableName')} target='_blank'>"
                           f"{item.attributes.get('tableName')}"
                           "</a>"
                          )
    html = ("<div class='item_container' style='height: auto; overflow: hidden; border: 1px solid #2BA4A0;"
            "border-radius: 2px; background: #2BA4A0; line-height: 1.21429em; padding: 10px;''>"
            "<div class='item_left' style='width: 210px; float: left;''>"
            f"{site_link}"
            f"{site_logo}"
            "</a></div><div class='item_right' style='float: none; width: auto; hidden;padding-left: 10px; overflow: hidden;''>"
            f"<a href={url_link} target='_blank'>"
            f"<b>{item.attributes.get('name')}</b>"
            "</a>"
            f"<br> {table_statement} | {kind_of_item} in {', '.join(item.attributes.get('application')).upper()}."
            f"<br>Last Modified: {item.attributes.get('updatedAt')}"
            f"<br><a href='{item.server}/v1/fields/{item.id}' target='_blank'>Fields</a>"
            f" Connector: {item.attributes.get('provider')}"
            f" | Published: {item.attributes.get('published')}"
            " </div> </div>")
    return html

def show_image_collection(item, i):
    html_string = ("<div class='item_container' style='height: auto; overflow: hidden; border: 1px solid #2BA4A0;"
                "border-radius: 2px; background: #2BA4A0; line-height: 1.21429em; padding: 10px;''>"
                "<div class='item_left' style='width: 100px; height: 100px; hidden;padding-left: 10px; float: left''>"
                f"<a href='{item.get('thumb_url')}' target='_blank'>"
                f"<img class='itemThumbnail' src='{item.get('thumb_url')}'>"
                "</a></div><div class='item_right' style='float: none; hidden;padding-left: 10px; width: auto; overflow: hidden;''>"
                f"<b>Image Source</b>: {item.get('instrument')} </br>"
                f"<b>Datetime</b>: {item.get('date_time')} </br>"
                f"<b>Cloud score </b>: {item.get('cloud_score')} </br>"
                " </div> </div>")
    return html_string

def show(item, i):
    """Returns an HTML block with template strings filled-in based on item attributes."""
    is_layer = item.get('type') == 'Layer'
    is_dataset = item.get('type') == 'Dataset'
    server = item.get('server', "https://api.skydipper.com")
    item_id = item.get('id', None)
    attributes = item.get('attributes', None)
    if is_layer:
        kind_of_item = 'Layer'
        url_link = f'{server}/v1/layer/{item_id}?includes=metadata'
    elif is_dataset:
        kind_of_item = 'Dataset'
        url_link = f'{server}/v1/dataset/{item_id}?includes=metadata,layer'
    else:
        kind_of_item = 'Unknown'
        url_link = None
    table_statement = f"Data source {attributes.get('provider')}"
    if attributes.get('connectorUrl') and attributes.get('provider') == "cartodb":
        table_statement = (f"Carto table: <a href={attributes.get('connectorUrl')}"
                           " target='_blank'>"
                           f"{attributes.get('tableName')}"
                           "</a>"
                          )
    if attributes.get('connectorUrl') and attributes.get('provider') == "csv":
        table_statement = (f"CSV Table: <a href={attributes.get('connectorUrl')}"
                           " target='_blank'>"
                           f"{attributes.get('tableName')}"
                           "</a>"
                          )
    if attributes.get('provider') == 'gee':
        table_statement = (f"GEE asset: <a href='https://code.earthengine.google.com/asset='"
                           f"{attributes.get('tableName')} target='_blank'>"
                           f"{attributes.get('tableName')}"
                           "</a>"
                          )
    site_link = "<a href='https://skydipper.com/' target='_blank'>"
    site_logo = "<img class='itemThumbnail' src='https://skydipper.com/images/logo.png'>"
    html = ("<div class='item_container' style='height: auto; overflow: hidden; border: 1px solid #2BA4A0;"
            "border-radius: 2px; background: #2BA4A0; line-height: 1.21429em; padding: 10px;''>"
            "<div class='item_left' style='width: 210px;  float: left;''>"
            f"{site_link}"
            f"{site_logo}"
            "</a></div><div class='item_right' style='float: none; width: auto; hidden;padding-left: 10px; overflow: hidden;''>"
            f"<b>{i}. </b>"
            f"<a href={url_link} target='_blank'>"
            f"<b>{attributes.get('name')}</b>"
            "</a>"
            f"<br> {table_statement} | {kind_of_item} in {', '.join(attributes.get('application')).upper()}."
            f"<br>Last Modified: {attributes.get('updatedAt')}"
            f"<br><a href='{server}/v1/fields/{item_id}' target='_blank'>Fields</a>"
            f" | Connector: {attributes.get('provider')}"
            f" | Published: {attributes.get('published')}"
            " </div> </div>")
    return html

def create_class(item):
    from .dataset import Dataset
    from .layer import Layer
    from .Skydipper import Widget
    from .image import Image
    if item['type'] == 'metadata':
        return Dataset(id_hash=item.get('attributes').get('dataset'), server = item.get('server'))
    if item['type'] == 'Dataset' :
        return Dataset(id_hash = item.get('id'), server = item.get('server'))
    elif item['type'] == 'Layer':
        return Layer(id_hash = item.get('id'), server = item.get('server'))
    elif item['type'] == 'Widget':
        return Widget(id_hash = item.get('id'), attributes=item.get('attributes'), server = item.get('server'))
    elif item['type'] == 'Image':
        return Image(**item)

def flatten_list(nested_list):
    if len(nested_list) > 0:
        return [item for sublist in nested_list for item in sublist]
    else:
        return []

def get_geojson_string(geom):
    coords = geom.get('coordinates', None)
    if coords and not any(isinstance(i, list) for i in coords[0]):
        geom['coordinates'] = [coords]
    feat_col = {"type": "FeatureCollection", "features": [{"type": "Feature", "properties": {}, "geometry": geom}]}
    return json.dumps(feat_col)

def parse_filters(filter_objects):
    filter_whitelist = ['connectorType', 'provider', 'status', 'published', 'protected', 'geoInfo']
    return_string = ''
    error_string = ''
    if filter_objects:
        for k, v in filter_objects.items():
            if k in filter_whitelist:
                return_string += f'{k}={v}&'
            else:
                error_string += f' {k},'
        if error_string:
            print(f'Unable to filter by{error_string[:-1]}.')
        return return_string
    return ''

def sldDump(sldObj):
    """
    Creates valid SldStyle string from an object.
    """
    sld_type = sldObj.get('type', None)
    extended = str(sldObj.get('extended', 'false')).lower()
    items = sldObj.get('items', None)
    sld_attr = {'color': None, 'label': None, 'quantity': None, 'opacity': None}

    if sld_type not in ['linear', 'ramp', 'gradient', 'intervals', 'values']:
        print('Unable to create SldStyle. Type must be in "linear", "ramp", "gradient", "intervals", "values".')
        return None
    if items:
        sld_str = f'<RasterSymbolizer> <ColorMap type="{sld_type}" extended="{extended}"> '
        for item in items:
            sld_str += f'<ColorMapEntry '
            for k in sld_attr.keys():
                if item.get(k, None): sld_str += f'{k}="{item[k]}" '
            sld_str += '/> + '
    return sld_str + "</ColorMap> </RasterSymbolizer>"

def sldParse(sld_str):
    """
    Builds a dictionary from an SldStyle string.
    """
    sld_str = sld_str.replace("'", '"').replace('\"', '"')
    keys = ['color', 'label', 'quantity', 'opacity']
    items = [el.strip() for el in sld_str.split('ColorMapEntry') if '<RasterSymbolizer>' not in el]
    sld_items = []
    for i in items:
        tmp = {}
        for k in keys:
            v = find_between(i, f'{k}="', '"')
            if v: tmp[k] = v
        sld_items.append(tmp)
    return {
        'type': find_between(sld_str, 'type="', '"'),
        'extended': find_between(sld_str, 'extended="', '"'),
        'items': sld_items
    }

def find_between(s, first, last):
    try:
        start = s.index(first) + len(first)
        end = s.index(last, start)
        return s[start:end]
    except ValueError:
        return ""

def nested_set(dic, keys, value):
    for key in keys[:-1]:
        dic = dic.setdefault(key, {})
    dic[keys[-1]] = value

def server_uses_widgets(server):
    """
    Does the server currently set use Widget objects? Response gives True if it does, false if not.
    """
    uses_widgets = ['https://api.resourcewatch.org','https://staging-api.globalforestwatch.org']
    if any(server in s for s in uses_widgets):
        return True
    else:
        return False

def tile_url(image, viz_params=None):
    """Create a target url for tiles from an EE image asset.
    e.g.
    im = ee.Image("LE7_TOA_1YEAR/" + year).select("B3","B2","B1")
    viz = {'opacity': 1, 'gain':3.5, 'bias':4, 'gamma':1.5}
    url = tile_url(image=im),viz_params=viz)
    """
    if viz_params:
        d = image.getMapId(viz_params)
    else:
        d = image.getMapId()
    base_url = 'https://earthengine.googleapis.com'
    url = (f"https://earthengine.googleapis.com/v1alpha/{d['mapid']}/tiles/{{z}}/{{x}}/{{y}}")
    return url


class EE_TILE_CALCS(object):
    """
    Copyright (c) 2018 Gennadii Donchyts. All rights reserved.
    This work is licensed under the terms of the MIT license.
    For a copy, see <https://opensource.org/licenses/MIT>.
    Refactored to Python, Vizzuality, 2020.
    This code will help you calculate tile bounds, intersected with
    a given geom, at a specified z-level.
    """
    def __init__(self, tileSize=256):
        ee.Initialize()
        self.tileSize = tileSize
        self.equatorial_circumference = 40075016.686
        self.origin = 2 * math.pi * 6378137 / 2.0

    def zoomToScale(self, zoom):
        tileWidth = self.equatorial_circumference / math.pow(2, zoom)
        pixelWidth = tileWidth / self.tileSize
        return pixelWidth

    def scaleToZoom(self, scale):
        tileWidth = scale * self.tileSize
        zoom = math.log(self.equatorial_circumference / tileWidth) / math.log(2)
        return math.ceil(zoom)

    def pixelsToMeters(self, px, py, zoom):
        resolution = self.zoomToScale(zoom)
        x = px * resolution - self.origin
        y = py * resolution - self.origin
        return [x, y]

    def metersToPixels(self, x, y, zoom):
        resolution = zoomToScale(zoom)
        px = (x + self.origin) / resolution
        py = (y + self.origin) / resolution
        return px, py

    def degreesToTiles(self, lon, lat, zoom):
        tx = math.floor((lon + 180) / 360 * math.pow(2, zoom))
        ty = math.floor((1 - math.log(math.tan(self.toRadians(lat)) + 1 / math.cos(self.toRadians(lat))) / math.pi) / 2 * math.pow(2, zoom))
        return [tx, ty]

    @staticmethod
    def tilesToDegrees(tx, ty, zoom):
        lon = tx / math.pow(2, zoom) * 360 - 180
        n = math.pi - 2 * math.pi * ty / math.pow(2, zoom)
        lat = toDegrees(math.atan(0.5 * (math.exp(n) - math.exp(-n))))
        return [lon, lat]

    def getTilesForGeometry(self, geometry, zoom):
        bounds = ee.List(geometry.bounds().coordinates().get(0))
        ll = bounds.get(0).getInfo() # <-- Look at making this happen server-side
        ur = bounds.get(2).getInfo() # <-- Look at making this happen server-side
        tmin = self.degreesToTiles(ll[0], ll[1], zoom)
        tmax = self.degreesToTiles(ur[0], ur[1], zoom)
        tiles = []
        for tx in range(tmin[0], tmax[0] + 1):
            for ty in range(tmax[1], tmin[1] + 1):
                bounds = self.getTileBounds(tx, ty, zoom)
                rect = ee.Geometry.Rectangle(bounds, 'EPSG:3857', False)
                tiles.append(ee.Feature(rect).set({'tx': tx, 'ty': ty, 'zoom': zoom }))
        return ee.FeatureCollection(tiles).filterBounds(geometry)

    def getTilesList(self, geometry, zoom):
        """Returns a list of individual features, where each feature element is a tile footprint."""
        bounds = ee.List(geometry.bounds().coordinates().get(0))
        ll = bounds.get(0).getInfo() # <-- Look at making this happen server-side
        ur = bounds.get(2).getInfo() # <-- Look at making this happen server-side
        tmin = self.degreesToTiles(ll[0], ll[1], zoom)
        tmax = self.degreesToTiles(ur[0], ur[1], zoom)
        tiles = []
        for tx in range(tmin[0], tmax[0] + 1):
            for ty in range(tmax[1], tmin[1] + 1):
                bounds = self.getTileBounds(tx, ty, zoom)
                rect = ee.Geometry.Rectangle(bounds, 'EPSG:3857', False)
                tiles.append(ee.Feature(rect).set({'tx': tx, 'ty': ty, 'zoom': zoom }))
        return tiles

    def getTileBounds(self, tx, ty, zoom, tileSize=256):
        """Returns a FeatureCollection object, where each feature is a tile footprint"""
        ty = math.pow(2, zoom) - ty - 1 # TMS -> XYZ, flip y index
        tmp_min = self.pixelsToMeters(tx * tileSize, ty * tileSize, zoom)
        tmp_max = self.pixelsToMeters((tx + 1) * tileSize, (ty + 1) * tileSize, zoom)
        return [tmp_min, tmp_max]

    @staticmethod
    def toRadians(degrees):
        return degrees * math.pi / 180

    @staticmethod
    def toDegrees(radians):
        return radians * 180 / math.pi


class MovieMaker(object):
    """
    Create movie tiles for a list of bounds, to go into a GCS bucket

    Parameters
    ----------
    area: ee.Geometry.Polygon()
        A polygon that covers the area which you want to generate tiles within.

    zlist: list
        A list of integer values of z-levels to process, e.g. z=[3] or z=[3,4,5]

    privatekey_path: string
        A string specifying the direction of a json keyfile on your local filesystem
        e.g. "/Users/me/.privateKeys/key_with_bucket_permissions.json"

    bucket_name: string
        A string specifying a GCS bucket (to which your private key should have access)
        e.g. 'skydipper_materials'

    folder_path: string
        A string specifying a folder name to create within the target bucket.
        e.g. 'movie-tiles/DTEST'

    report_status: bool
        Set to true if you want the program to report on the files it is generating.
        Beware it can get long for high z-levels.
    """
    def __init__(self, privatekey_path, bucket_name, folder_path,
                    area=None, zlist=None, ic=None, report_status=False):
        self.storage_client = storage.Client.from_service_account_json(privatekey_path)
        self.privatekey_path = privatekey_path
        self.bucket = self.storage_client.get_bucket(bucket_name)
        self.bucket_name = bucket_name
        self.folder_path = folder_path
        self.tiler = EE_TILE_CALCS()
        self.area = area
        self.zlist = zlist
        self.ic = ic
        self.report_status = report_status
        ee.Initialize()

    def run(self):
        """Main worker method"""
        assert type(self.zlist) == list, "the zlist must be a list to run, e.g. zlist=[2]"
        assert type(self.area) == ee.geometry.Geometry, "An area of type ee.geometry.Geometry must be provided to run"
        for zlevel in self.zlist:
            print(f"🧨 Calculating Z-level {zlevel}")
            tileset = self.tiler.getTilesList(self.area, zlevel)
            d = self.initial_dic_creation(tileset=tileset) # Starting dic of whatever has been burned to the bucket
            to_do = self.get_items_by_state(d, 'WAITING')
            for unprocessed in to_do:
                z=unprocessed[0].split('/')[-3]
                x=unprocessed[0].split('/')[-2]
                y=unprocessed[0].split('/')[-1].split('.mp4')[0]
                if self.report_status: print(f'{z}/{x}/{y}')
                try:
                    self.movie_maker(tile=unprocessed[1].get('tile'), z=z, x=x, y=y)
                except (ee.EEException) as err:
                    sleep(60 * 5)  # Simple - Wait 5 mins and try assigning tasks again (this assumes the only issue )
                    self.movie_maker(tile=unprocessed[1].get('tile'), z=z, x=x, y=y)
        print("Program ended normally. Note that after the files have been generated you should run MovieMaker().reNamer()")
        self.reNamer()
        return

    def movie_maker(self, tile, z, x, y):
        """Generates a single movie tile"""
        g = tile.geometry()
        filtered = self.ic.filterBounds(g)
        #print(f"🗺 Exporting movie-tile to {self.bucket_name}/{self.folder_path}/{z}/{x}/{y}.mp4")
        exp_task = ee.batch.Export.video.toCloudStorage(
                        collection = filtered,
                        description = f'{z}_{x}_{y}',
                        bucket= self.bucket_name,
                        fileNamePrefix = f"{self.folder_path}/{z}/{x}/{y}",
                        dimensions = [256,256],
                        framesPerSecond = 2,
                        region = g)
        exp_task.start()

    def reNamer(self):
        """Renames source files to a clean target that removes jank added by EE."""
        blob_gen = self.bucket.list_blobs(prefix=self.folder_path)
        blobs = [blob for blob in blob_gen]
        print(f'Scanning target {self.bucket_name}/{self.folder_path} for files that require renaming...')
        for blob in blobs:
            tmp_name = blob.name
            if tmp_name[-4:] == '.mp4' and ("ee-export-video" in tmp_name):
                target_name = f"{tmp_name.split('ee-export-video')[0]}.mp4"
                if self.report_status: print(f'renaming:{tmp_name}-->{target_name}')
                _ = self.bucket.rename_blob(blob, target_name)

    def getDoneFileList(self):
        """Returns list of file names in a bucket/path that have already been created"""
        blob_gen = self.bucket.list_blobs(prefix=self.folder_path)
        blobs = [blob for blob in blob_gen]
        completed_files = []
        for blob in blobs:
            completed_files.append(blob.name)
        return completed_files

    def getFullTargetList(self, z):
        """
            Return a list of all files we intend to create for a specified z (int) target to cover a given ee.Geometry input
            area, with a folder path (string) direction appended.
        """
        target_list = []
        tmp_fc = self.tiler.getTilesForGeometry(self.area, z)
        tmp_info = tmp_fc.getInfo()   # <---this is the point where I have the properties, I should make sure to propagate them rather then call getinfo again
        target_list = [f"{self.folder_path}/{item.get('properties').get('zoom')}/{item.get('properties').get('tx')}/{item.get('properties').get('ty')}.mp4"
                       for item in tmp_info.get('features')]
        return sorted(target_list)

    def get_current_status(self):
        """Consult the current EE Task list to see what's what"""
        batch_jobs = ee.batch.Task.list()
        processing_list = []
        processing_status = []
        badness_list = []
        for job in batch_jobs:
            state = job.state
            try:
                tmp = job.config['description'].split("_")
                tmp_fname = f"{self.folder_path}/{tmp[0]}/{tmp[1]}/{tmp[2]}.mp4"
                processing_list.append(tmp_fname)
                processing_status.append(state)
            except:
                badness_list.append(job)
        return {'processing':processing_list,
                'status': processing_status,
                'badness': badness_list}

    def get_objective_list(self, tileset):
        """Returns a list of target files 1:1 for the target tiles"""
        tmp_list = []
        for tile in tileset:
            tmp_str = tile.__str__()
            zoom = json.loads(tmp_str[11:-1]).get('arguments').get('value')
            ty = json.loads(tmp_str[11:-1]).get('arguments').get('object').get('arguments').get('value')
            tx = json.loads(tmp_str[11:-1]).get('arguments').get('object').get('arguments').get('object').get('arguments').get('value')
            tmp_list.append(f"{self.folder_path}/{zoom}/{tx}/{ty}.mp4")
        return tmp_list

    @staticmethod
    def generate_master_dic(objectives, tileset):
        d = {}
        for obj, tile in zip(objectives, tileset):
            d[obj] = {'tile': tile, 'status': "WAITING"}
        return d

    @staticmethod
    def get_items_by_state(d, state="WAITING"):
        result = []
        for i in d.items():
            if (i[1].get('status') == state):
                result.append(i)
        return result


    def initial_dic_creation(self, tileset):
        objectives = self.get_objective_list(tileset) # <--- calculate target files to keep track
        d = self.generate_master_dic(objectives=objectives, tileset=tileset)
        self.reNamer()
        done = self.getDoneFileList() # Consult the bucket and see what files have been completed
        for item in done:
            if d.get(item, None):
                d[item]['status']='COMPLETED'
        return d


import requests
import random
import os
import json
import datetime
from tqdm import tqdm
from .dataset import Dataset
from .layer import Layer
from .utils import create_class, show, flatten_list, parse_filters
from .user import User
from .Skydipper import Metadata

class Collection:
    """
    Returns a list of objects from a server

    This function searches all avaiable dataset entries within user specified limits and returns a list.
    of objects.

    Parameters
    ----------
    name: str
        A string to search for
    altname: str
        An alternative name string to search for
    description: str
        A description string to search for
    citation: str
        A citation string to search for
    language: str
        A two-character language code string to search for (e.g. 'en')
    app: list
        A list of string IDs of applications to search, e.g. ['skydipper', ‘soilsRevealed’]
    limit: int
        Maximum number of items to return
    order: str
        Field to order items by, e.g. ’date’
    sort: str
        Rule to sort items by, either ascending (’asc’) or descending ('desc')
    search: str
        String to search records by, e.g. ’Forest loss’
    filters: dict
        A dictionary of filter key, value pairs e.g. {'provider', 'gee'}
        Possible search keys: 'connectorType', 'provider', 'status', 'published', 'protected', 'geoInfo'.
    """
    def __init__(self, name=None, altname=None, description=None, app=['skydipper','soilsRevealed','test'], env='production', limit=1000, order='name', sort='desc',
                server="https://api.skydipper.com", language=None, citation=None,
                 filters=None, mapbox_token=None):
        self.User = User()
        # self.search = search
        self.name = name
        self.altname = altname
        self.description = description
        self.server = server
        self.app = ",".join(app)
        self.env = env
        self.limit = limit
        self.language = language
        self.citation = citation
        self.order = order
        self.sort = sort
        self.filters = filters
        self.mapbox_token = mapbox_token
        self.object_type = ['datasett']
        self.payload = self.get_payload()
        self.metadata = self.get_metadata()
        #self.collection = self.get_collection()
        self.iter_position = 0

    def _repr_html_(self):
        str_html = ""
        for n, c in enumerate(self.metadata):
            c['server'] = self.server
            str_html += show(c, n)
            if n < len(self.metadata)-1:
                str_html += '<p></p>'
        return str_html

    def __repr__(self):
        rep_string = "["
        for n, c in enumerate(self.metadata):
            rep_string += str(f"{n}. {c.get('type')} {c.get('id')} {c.get('attributes', None).get('name', None)}")
            if n < len(self.metadata)-1:
                rep_string += ',\n '
        rep_string += ']'
        return rep_string

    def __iter__(self):
        return self

    def __next__(self):
        if self.iter_position >= len(self.metadata):
            self.iter_position = 0
            raise StopIteration
        else:
            self.iter_position += 1
            return self.metadata[self.iter_position - 1]

    def __getitem__(self, key):
        items = self.metadata[key]
        if type(items) == list:
            return [create_class(item) for item in items]
        else:
            return create_class(items)

    def __len__(self):
        return len(self.metadata)

    def get_collection(self):
        """
        Getter for the a collection object. In this case dataset and layers
        are the objects in the API. I.e. tables are a dataset type.
        """
        metadata = self.get_metadata()
        id_hashes = [item.get('attributes').get('dataset') for item in metadata if len(metadata) > 0]
        for id_hash in id_hashes:
            collection=[]
            try:
                collection.append(Dataset(id_hash=id_hash, server=self.server))
            except:
                pass
        return collection

    def get_payload(self):
        payload = {}
        if self.name: payload['name'] = self.name
        if self.altname: payload['altName'] = self.altname
        if self.description: payload['description'] = self.description
        if self.app: payload['app'] = self.app
        if self.language: payload['language'] = self.language
        if self.citation: payload['citation'] = self.citation
        return payload

    def get_metadata(self):
        url = f"{self.server}/v1/search"
        r = requests.get(url, params=self.payload, headers=self.User.headers)
        response_list = r.json().get('data', None)
        if not response_list:
            raise ValueError('No items found')
        #return [Metadata(attributes=item, server=self.server) for item in response_list]
        # in the future changing this to metadata native type is the right thing to do
        return response_list

    # def filter_results(self, response_list):
    #     """Search by a list of strings to return a filtered list of Dataset or Layer objects"""
    #     filtered_response = []
    #     collection = []
    #     for item in response_list:
    #         return_layers = 'layer' in self.object_type
    #         return_datasets = 'dataset' in self.object_type
    #         return_tables = 'dataset' in self.object_type
    #         #print(f"Filter: {type(item)}, {item}")
    #         if type(item) == dict:
    #             tmp_atts = item.get('attributes')
    #             name = tmp_atts.get('name', None)
    #             description = tmp_atts.get('description', None)
    #             slug = tmp_atts.get('slug', None)
    #         else:
    #             name = None
    #             description = None
    #             slug = None
    #         found = []
    #         if description:
    #             description = description.lower()
    #             found.append(any([s in description for s in self.search]))
    #         if name:
    #             name = name.lower()
    #             found.append(any([s in name for s in self.search]))
    #         if slug:
    #             slug = slug.lower().split('_')
    #             found.append(any([s in slug for s in self.search]))
    #         if any(found):
    #             if len(filtered_response) < self.limit:
    #                 filtered_response.append(item)
    #             elif item.get('type') == 'dataset' and item.get('attributes').get('provider') != ['csv','json'] and return_datasets:
    #                 collection.append({'type': 'Dataset','id': item.get('id'), 'attributes': item.get('attributes'), 'server': self.server})
    #             if item.get('type') == 'layer' and return_layers:
    #                 collection.append({'type': 'Layer', 'id': item.get('id'), 'attributes': item.get('attributes'), 'server': self.server, 'mapbox_token':self.mapbox_token})
    #     return collection

    # def order_results(self, collection_list):
    #     """Operate on a list of objects given the order key, limit, and rule a user has passed"""
    #     tmp_sorted = []
    #     try:
    #         d = {}
    #         duplicate = {}
    #         for n, z in enumerate([c['attributes'].get(self.order.lower()) for c in collection_list]):
    #             if duplicate.get(z, None):
    #                 d[f'{z} {duplicate[z]}'] = collection_list[n]
    #                 duplicate[z] += 1
    #             else:
    #                 d[z] = collection_list[n]
    #                 duplicate[z] = 1
    #         keys = sorted(d, reverse=self.sort.lower() == 'asc')
    #         for key in keys:
    #             tmp_sorted.append(d[key])
    #     except:
    #         raise ValueError(f'[Order-error] Param does not exist in collection: {self.order}, rule: {self.sort}')
    #     if self.limit < len(tmp_sorted):
    #         tmp_sorted = tmp_sorted[0:self.limit]
    #     return tmp_sorted

    def save(self, path=None):
        """
        Save all entities in the collection to a local path.
        """
        if not path:
            path = './LMI-BACKUP'
            if not os.path.isdir(path):
                os.mkdir(path)
            today = datetime.datetime.today().strftime('%Y-%m-%d@%Hh-%Mm')
            path += f'/{today}'
            if not os.path.isdir(path):
                os.mkdir(path)
        else:
           if not os.path.isdir(path):
                os.mkdir(path)
        print(f'Saving to path: {path}')
        saved = []
        failed = []
        for item in tqdm(self):
            if item['id'] not in saved:
                entity_type = item.get('type')
                if entity_type in ['Dataset']:
                    ds_id = item['id']
                else:
                    ds_id = item['attributes']['dataset']
                try:
                    url = f'{self.server}/v1/dataset/{ds_id}?includes=metadata,layer'
                    r = requests.get(url, headers=self.User.headers)
                    dataset_config = r.json()['data']
                except:
                    failed.append(item)

                save_json = {
                    "id": ds_id,
                    "type": "dataset",
                    "server": self.server,
                    "attributes": dataset_config['attributes']
                }
                with open(f"{path}/{ds_id}.json", 'w') as fp:
                    json.dump(save_json, fp)

        if len(failed) > 0:
            print(f'Some entities failed to save: {failed}')
            return failed
        print('Save complete!')



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="Skydipper",
    version="0.2.0",
    author="Vizzuality",
    author_email="benjamin.laken@vizzuality.com",
    description="Pythonic interface to the Skydipper API.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    license='MIT',
    url="https://github.com/Skydipper/Skydipper",
    install_requires=['requests>=2.2.0',
                        'folium==0.8.3',
                        'google-cloud-storage==1.25.0',
                        'earthengine-api==0.1.210',
                        'geopandas>=0.4.1',
                        'geojson>=2.4.0',
                        'pypng>=0.0.19',
                        'tqdm==4.41.1'],
    packages=['Skydipper'],
    classifiers=[
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)



from .Skydipper import Vocabulary, Metadata, Widget
from .layer import Layer
from .image import Image
from .user import User
from .imageCollection import ImageCollection
from .dataset import Dataset
from .geometry import Geometry
from .collection import Collection
from pkg_resources import get_distribution
import os
import requests

__version__ = get_distribution('Skydipper').version